/*
 * Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.binary;

import com.oracle.truffle.api.dsl.Fallback;
import com.oracle.truffle.api.dsl.NodeChild;
import com.oracle.truffle.api.dsl.Specialization;
import com.oracle.truffle.api.dsl.TypeSystemReference;
import com.oracle.truffle.r.nodes.EmptyTypeSystemFlatLayout;
import com.oracle.truffle.r.runtime.RRuntime;
import com.oracle.truffle.r.runtime.builtins.RSpecialFactory;
import com.oracle.truffle.r.runtime.nodes.RNode;
import com.oracle.truffle.r.runtime.ops.BooleanOperation;
import com.oracle.truffle.r.runtime.ops.BooleanOperationFactory;

/**
 * Fast-path for scalar values: these cannot have any class attribute. Note: we intentionally use
 * empty type system to avoid conversions to vector types. NA values cause
 * {@link RSpecialFactory#throwFullCallNeeded()} exception.
 */
@TypeSystemReference(EmptyTypeSystemFlatLayout.class)
@NodeChild(value = "arguments", type = RNode[].class)
public abstract class BinaryBooleanSpecial extends RNode {
    @Child private BooleanOperation operation;

    protected BinaryBooleanSpecial(BooleanOperation operation) {
        this.operation = operation;
    }

    public static RSpecialFactory createSpecialFactory(final BooleanOperationFactory opFactory) {
        return (signature, arguments) -> signature.getNonNullCount() == 0 && arguments.length == 2 ? BinaryBooleanSpecialNodeGen.create(opFactory.create(), arguments) : null;
    }

    @Specialization
    public byte doInts(int left, int right) {
        if (RRuntime.isNA(left) || RRuntime.isNA(right)) {
            throw RSpecialFactory.throwFullCallNeeded();
        }
        return RRuntime.asLogical(operation.op(left, right));
    }

    @Specialization
    public byte doDoubles(double left, double right) {
        if (!RRuntime.isFinite(left) || !RRuntime.isFinite(right)) {
            throw RSpecialFactory.throwFullCallNeeded();
        }
        return RRuntime.asLogical(operation.op(left, right));
    }

    @Specialization
    public byte doIntDouble(int left, double right) {
        if (RRuntime.isNA(left) || !RRuntime.isFinite(right)) {
            throw RSpecialFactory.throwFullCallNeeded();
        }
        return RRuntime.asLogical(operation.op(left, right));
    }

    @Specialization
    public byte doDoubleInt(double left, int right) {
        if (!RRuntime.isFinite(left) || RRuntime.isNA(right)) {
            throw RSpecialFactory.throwFullCallNeeded();
        }
        return RRuntime.asLogical(operation.op(left, right));
    }

    @Fallback
    @SuppressWarnings("unused")
    public void doFallback(Object left, Object right) {
        throw RSpecialFactory.throwFullCallNeeded();
    }

}
